      	program poisson

!===================================================================================
!  14 Apr 98
!  Paul M. Alsing
!  Albuqueruqe High Performance Computing Center
!  1601 Central NE
!  University of New Mexico
!  Albuquerque, New Mexico 87131
!  alsing@arc.unm.edu
!  505-277-9094
!===================================================================================
! Iteratively solve Poisson Equation 
!             d^2phi(x,y)/dx^2 + d^2phi(x,y)/dy^2 = -4*pi*rho(x,y)
! for the potential phi(x,y) given the charge density rho(x,y) on a N*M point
! grid of physical length Lx by Ly on each side. Solve with a rectangular processor 
! grid of size np_rows and np_cols.  Use MPI Cartesian Topology functions
! to help with the parallel communications of boundary elements amongst processors. 
!===================================================================================

	use poisson_routines
      	use string_utilities

      	implicit none


      	integer 			:: N, M
	integer				:: i,j, it

	real(DP)			:: Lx, Ly, dx, dy
      	real(DP) 			:: norm_local,norm_global,sum_rho_local,sum_rho_global
      	real(DP), allocatable 		:: rho_local(:,:), phi_local(:,:), phi_old_local(:,:)     
      	real(DP), parameter 		:: eps = 1d-5
   	real(DP)			:: rtc   !for AIX machines only
	real(DP)			:: t_start, t_end, t_exec_max
	real(DP)			:: sum_phi_local, sum_phi_global
 	real(DP)			:: f, fi, fj				!scale factors
	real(DP), allocatable		:: right_ghost_cells(:)
	real(DP), allocatable		:: left_ghost_cells(:)
	real(DP), allocatable		:: top_ghost_cells(:)
	real(DP), allocatable		:: bottom_ghost_cells(:)
	real(DP), allocatable		:: right_left_buffer(:)
	real(DP), allocatable		:: top_bottom_buffer(:)

! MPI Variables

	integer				:: rank		! rank in MPI_COMM_WORLD
	integer				:: tag=100	! communicator tag (arbitrary value)
	integer				:: ierr
	integer				:: NP		! total # procrs used
	integer				:: n_extra
	integer				:: status(MPI_STATUS_SIZE)	
	integer				:: int_data(10)
	real(DP)			:: dble_data(10)


	! grid variables

	integer				:: number_of_dims
	integer				:: dim_sizes(2)
	logical				:: wrap_around(2)	!0->no, 1->yes
	logical				:: reorder		!0->no, 1->yes
	logical				:: free_coords(2)	! used to create the sub communicators

	integer				:: rank_recv_from
	integer				:: rank_send_to
	integer				:: direction		!0->i varies(ie N,S), 1->j varies(ie E,W)
	integer				:: disp			!# procr to displace in a shift

! final output arrays

	integer, allocatable		:: recvcounts1(:), recvcounts2(:)
	integer, allocatable		:: displs1(:),    displs2(:)
	real(DP),allocatable		:: recvbuf1(:,:), Trecvbuf1(:,:), Tphi_global(:,:)
 
! final output arrays for rho the charge density

	real(DP),allocatable		:: recvbuf1_rho(:,:), Trecvbuf1_rho(:,:), Tphi_global_rho(:,:)


! Initialize MPI

	call MPI_INIT( ierr )
	call MPI_COMM_SIZE(MPI_COMM_WORLD, NP,   ierr)
	call MPI_COMM_RANK(MPI_COMM_WORLD, rank, ierr)

	print'(a,I2,a)', 'P(',rank,'): saying hello'
	if(rank==0) print'(2(a,I2))', 'P(',rank,'): Using # procrs, NP = ',NP

!-----------------------------------------------------------------------------
!				User Inputs
!-----------------------------------------------------------------------------	

! User inputs

user:	if(rank==0) then
	   
	   print'(/,a)','Input the # of rows and cols in the procr grid'
	   read*, np_rows, np_cols
	   print*,'Using np_rows = ',np_rows,'  np_cols = ',np_cols

      	   print'(/,a)', 'Input the length of each side of the grid Lx, Ly'
	   print*,       'and the # of grid points N, M in each x and y directions'
      	   read*, Lx, Ly, N, M
	   print*,'Solving Poisson Equation for Lx = ',real(Lx),'  Ly = ',real(Ly), &
	          ' N = ',N,' M = ',M
	
! load up information to send to other procrs

	   int_data(1) = np_rows
	   int_data(2) = np_cols
	   int_data(3) = N
	   int_data(4) = M

	   dble_data(1) = Lx
	   dble_data(2) = Ly

	endif user

!  broadcast user inputs to all procrs

	call MPI_BCAST( int_data,  10, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr )
	call MPI_BCAST( dble_data, 10, MPI_DOUBLE_PRECISION,  0, MPI_COMM_WORLD, ierr )	


	np_rows = int_data(1)  
	np_cols = int_data(2)  
	N 	= int_data(3) 
	M 	= int_data(4) 

	Lx	= dble_data(1) 
	Ly	= dble_data(2)


!  check for consistent procr grid

	if( NP /=  np_rows*np_cols ) then
	    if(rank==0) then
	       print*,'***ERROR***'
	       print*,'# procrs used NP = ',NP,' must equal np_rows*np_cols'
	    endif
	    call MPI_FINALIZE(ierr)
	    stop
	endif 

!-----------------------------------------------------------------------------
!				Create Communicators
!-----------------------------------------------------------------------------	

!  create the grid communicator

	number_of_dims = 2
	dim_sizes(1)   = np_rows
	dim_sizes(2)   = np_cols
	wrap_around    = .true.		! cylic procr grid in both directions
	reorder	       = .false.	! keep same procr ordering (row-major) 
					! as in MPI_COMM_WORLD

	call MPI_CART_CREATE( MPI_COMM_WORLD, number_of_dims, dim_sizes, wrap_around, 	&
			                      reorder, grid_comm, ierr)


!  create row sub-communicators from grid_comm

	free_coords(1) = .false.	!i.e  do not allow the procr rows to vary
	free_coords(2) = .true.		!i.e. allow  the procr columns to vary

	call MPI_CART_SUB( grid_comm, free_coords, row_comm, ierr )


!  create column sub-communicators from grid_comm

	free_coords(1) = .true.		!i.e  allow  the procr rows to vary
	free_coords(2) = .false.	!i.e. do not allow the procr columns to vary

	call MPI_CART_SUB( grid_comm, free_coords, col_comm, ierr )


!  determine the # of rows and columns in the local matrices
 
	allocate( nmbr_local_rows(np_rows), nmbr_local_cols(np_cols) )

	nmbr_local_rows = N/np_rows
	n_extra = modulo( N, np_rows )
	do i =  1, n_extra
           nmbr_local_rows(i) = nmbr_local_rows(i)  + 1
	enddo

	nmbr_local_cols = M/np_cols
	n_extra = modulo( M, np_cols )
	do j =  1, n_extra
           nmbr_local_cols(j) = nmbr_local_cols(j)  + 1
	enddo

!  find out who you are
!  NOTE: MPI uses 0-based coordintate i.e.
!        0 <= coord(1) <= np_rows-1
!        0 <= coord(2) <= np_cols-1
!   
!	 Since were using 1-based arrays we have to add +1 to each of these coordinates

	call MPI_COMM_RANK( grid_comm, grid_rank, ierr )
	call MPI_CART_COORDS( grid_comm, grid_rank, number_of_dims, coords, ierr )	
 
	N_local = nmbr_local_rows( coords(1) + 1 )	! coords are 0-based; nmbr_local_rows 1-based
	M_local = nmbr_local_cols( coords(2) + 1 )	! coords are 0-based; nmbr_local_cols 1-based


!-----------------------------------------------------------------------------
!			Allocate memory and initialize
!-----------------------------------------------------------------------------	

!  allocate memory for arrays
!  include ghost cells for boundary data from nearby procrs   

      	allocate ( rho_local(N_local,M_local) )			!charge density
      	allocate ( phi_local(0:N_local+1,0:M_local+1) )		!potential
      	allocate ( phi_old_local(0:N_local+1,0:M_local+1) )	!potential, previous iteration

	allocate( right_ghost_cells(0:N_local+1),  left_ghost_cells(0:N_local+1)   )
	allocate( top_ghost_cells(0:M_local+1)  ,  bottom_ghost_cells(0:M_local+1) )
	allocate( right_left_buffer(0:N_local+1),  top_bottom_buffer(0:M_local+1) )


! initialization

      	dx=Lx/N; dy=Ly/M
	f  = 2*pi*(dx*dy)**2 / (dx*dx+dy*dy)		!scale factor for rho_local
	fi = .5*dy*dy / (dx*dx+dy*dy)			!scale factor in discretization
	fj = .5*dx*dx / (dx*dx+dy*dy) 			!scale factor in discretization

      	rho_local = makerho_local(Lx,Ly,N,M)*f		!initial charge density

      	phi_local = 0.0					!set potential to 0
      	sum_rho_local  = sum(abs(rho_local))		!sum of abs(rho) over local grid

	call MPI_ALLREDUCE(sum_rho_local, sum_rho_global, 1, MPI_DOUBLE_PRECISION, MPI_SUM,  &
			   			             MPI_COMM_WORLD, ierr )



!-----------------------------------------------------------------------------
!-----------------------------------------------------------------------------
!				Main Compuational Loop
!-----------------------------------------------------------------------------
!-----------------------------------------------------------------------------	

!  Iteratively solve Poisson Equation

	call MPI_BARRIER( MPI_COMM_WORLD, ierr )
	if(rank==0) then
	    print*,' '
            print*,'Beginning poisson solution for d^2phi(x,y)/dx^2 + d^2phi(x,y)/dy^2 = -4 pi rho(x,y)'
            print*,'Solution will converge when |phi_new - phi_old| < ',eps
	    print*,' '
	endif 
	
	it = 0

  	t_start = MPI_WTIME()   !rtc() - for AIX only

iterate: do 

	   it = it + 1

           phi_old_local=phi_local			!save old value of phi


!-----------------------------------------------------------------------------
!			Parallel Communication Step
!-----------------------------------------------------------------------------

!  perform 4 cyclic shifts on the procr grid, N,S,E,W  in order to bring
!  boundary data into ghost cells
!  NOTE: In the following (direction=0, direction=1) corresponds to  (i,j)  
!	A call to MPI_CART_SHIFT with direction=0 means the i index of P(i,j) is varying 
!	i.e a column shift N(up) or S(down), if disp<0, disp>0, respectively. 
!	A call to MPI_CART_SHIFT with direction=1 means the j index of P(i,j) is varying 
!	i.e a row shift E or W if disp>0, disp<0, respectively.


!  Load up the ghost cells

!-----------------------------------------------------------------------------
!  N or upward   shift:  P(i+1,j) ->(recv from)-> P(i,j) ->(send to)-> P(i-1,j)
!-----------------------------------------------------------------------------
	direction =  0	!i varies
	disp      = -1	!i->i-1
	top_bottom_buffer = phi_old_local(1,:)		!sending data, row 1
	call MPI_CART_SHIFT( grid_comm, direction, disp, rank_recv_from, rank_send_to, ierr)	
	call MPI_SENDRECV( top_bottom_buffer,  M_local+2, MPI_DOUBLE_PRECISION, rank_send_to,   tag, &
			   bottom_ghost_cells, M_local+2, MPI_DOUBLE_PRECISION, rank_recv_from, tag, &
			   grid_comm, status, ierr )
	phi_old_local(N_local+1,:) = bottom_ghost_cells;
!print*,'N shift: sending to ',rank_send_to,'; recving from ',rank_recv_from
!print*,'Nsend: ', top_bottom_buffer
!print*,'Nrecv: ', bottom_ghost_cells

!-----------------------------------------------------------------------------
!  S or downward shift:  P(i-1,j) ->(recv from)-> P(i,j) ->(send to)-> P(i+1,j)
!-----------------------------------------------------------------------------
	direction =  0	!i varies
	disp      =  1	!i->i+1
	top_bottom_buffer = phi_old_local(N_local,:)	!sending data, row N_local
	call MPI_CART_SHIFT( grid_comm, direction, disp, rank_recv_from, rank_send_to, ierr)
	call MPI_SENDRECV( top_bottom_buffer,  M_local+2, MPI_DOUBLE_PRECISION, rank_send_to,   tag, &
			   top_ghost_cells,    M_local+2, MPI_DOUBLE_PRECISION, rank_recv_from, tag, &
			   grid_comm, status, ierr )
	phi_old_local(0,:) = top_ghost_cells
!print*,'S shift: sending to ',rank_send_to,'; recving from ',rank_recv_from
!print*,'Ssend: ', top_bottom_buffer
!print*,'Srecv: ', top_ghost_cells 


!----------------------------------------------------------------------------- 
!  E or right    shift:  P(i,j-1) ->(recv from)-> P(i,j) ->(send to)-> P(i,j+1)
!-----------------------------------------------------------------------------
	direction =  1	!j varies
	disp      =  1	!j->j+1
	right_left_buffer = phi_old_local(:,M_local)	!sending data, col M_local
	call MPI_CART_SHIFT( grid_comm, direction, disp, rank_recv_from, rank_send_to, ierr)
	call MPI_SENDRECV( right_left_buffer,  N_local+2, MPI_DOUBLE_PRECISION, rank_send_to,   tag, &
			   left_ghost_cells,   N_local+2, MPI_DOUBLE_PRECISION, rank_recv_from, tag, &
			   grid_comm, status, ierr )
	phi_old_local(:,0) = left_ghost_cells
!print*,'E shift: sending to ',rank_send_to,'; recving from ',rank_recv_from
!print*,'Esend: ', right_left_buffer
!print*,'Erecv: ', left_ghost_cells

!-----------------------------------------------------------------------------
!  W or left    shift:   P(i,j+1) ->(recv from)-> P(i,j) ->(send to)-> P(i,j-1)
!-----------------------------------------------------------------------------
	direction =  1	!j varies
	disp      = -1	!j->j-1
	right_left_buffer = phi_old_local(:,1)		!sending data, col 1
	call MPI_CART_SHIFT( grid_comm, direction, disp, rank_recv_from, rank_send_to, ierr)
	call MPI_SENDRECV( right_left_buffer,  N_local+2, MPI_DOUBLE_PRECISION, rank_send_to,   tag, &
			   right_ghost_cells,  N_local+2, MPI_DOUBLE_PRECISION, rank_recv_from, tag, &
			   grid_comm, status, ierr )
	phi_old_local(:,M_local+1) = right_ghost_cells
!print*,'W shift: sending to ',rank_send_to,'; recving from ',rank_recv_from
!print*,'Wsend: ', right_left_buffer
!print*,'Wrecv: ', right_ghost_cells


!-----------------------------------------------------------------------------
!  NOTE: phi_old_local(0:N_local+1,0:M_local+1) now contains all the data 
!        necessary on procr to perform the local computation
!-----------------------------------------------------------------------------

!diagnostics
if(it==1) then
open( 2, file='phi_1it.'//char(48+coords(1))//char(48+coords(2)) )
do i=0,N_local+1
   do j = 0, M_local+1
      write(2,'(F3.1,2x)',advance='no') real( phi_old_local(i,j) )
   enddo
   write(2,*) ' '
enddo
close(2)
endif


!-----------------------------------------------------------------------------
!			Local (on Procr) Computation
!-----------------------------------------------------------------------------

!  local computation of phi on each procr
!  Note: indices say within range because of the boundary ghostcells


	       do j = 1, M_local
	          do  i = 1, N_local
	              phi_local(i,j) = rho_local(i,j) 					   &
		      	    + fi * ( phi_old_local( i+1, j ) +  phi_old_local( i-1, j ) )  &			
		   	    + fj * ( phi_old_local( i, j+1 ) +  phi_old_local( i, j-1 ) ) 	
		  enddo
	       enddo


!-----------------------------------------------------------------------------
!		Parallel Communication to determine stopping criteria
!-----------------------------------------------------------------------------

	   ! The solution phi is undetermined up to an arbitray constant.
	   ! We fix this constant by subtracting off, at each iteration,
	   ! the average value of phi over the whole grid 


	   sum_phi_local = SUM(phi_local(1:N_local,1:M_local))
	 
	   call MPI_ALLREDUCE(sum_phi_local, sum_phi_global, 1, MPI_DOUBLE_PRECISION, MPI_SUM,  &
			   				        MPI_COMM_WORLD, ierr )

           !subtract out average value of phi
           phi_local(1:N_local,1:M_local)=phi_local(1:N_local,1:M_local) - sum_phi_global/N/M	



	   ! compute norm of difference between current and old solution

	   !local norm
           norm_local = SUM( abs(phi_local(1:N_local,1:M_local)-phi_old_local(1:N_local,1:M_local)) ) / sum_rho_global	


	   call MPI_ALLREDUCE(norm_local, norm_global, 1, MPI_DOUBLE_PRECISION, MPI_SUM,  &
			   			          MPI_COMM_WORLD, ierr )


	   if(norm_global < eps) exit


	   ! test for non-convergence

	   if( it >= ITMAX ) then
	       if(rank==0) print*,'*** Solution failed to converger after ',ITMAX,' iterations'
	       call MPI_FINALIZE( ierr )
	       stop 'EXITING'
	   endif

! diagnostic output

	if( modulo(it, NOUTPUT) == 0 .and. rank==0 ) then
	    print*,'P(0): At it = ',it,' norm = ',norm_global
	endif

	enddo iterate


!  final output message

	
	if(rank==0)then
	print*,'Poisson calculation complete.'
	   print*,'P(',rank,'): At it = ',it,' norm = ',real(norm_global),'  convergence!'
        endif
  	t_end = MPI_WTIME()  !rtc() - for AIX only
	print*,'P(',rank,'): Execution time = ',t_end-t_start,' secs for ',it,	 &
	     & ' iterations, or ',(t_end-t_start)/it,' secs per iteration. N = ',N
	call MPI_REDUCE( t_end-t_start, t_exec_max, 1, MPI_DOUBLE_PRECISION, MPI_MAX,  & 
						   0, MPI_COMM_WORLD, ierr )
	if(rank==0) print*,'P(',rank,'): Total Execution time across all procrs = ',t_exec_max,' secs '


!-----------------------------------------------------------------------------
!		Parallel Operations for final output
!-----------------------------------------------------------------------------
! The following is one possible strategy for outputing the global array phi.
! This violates the parallel domain decomposition strategy of no one processor
! every having to allocate memory for the entire global array. However, this
! exercise illustrates the use of MPI_GATHERV and row communicators
! to obtain the entire global array in the correct fortran column-oriented
! order. Note: another possible way to output the array would be for
! each processor to output its local array, phi_local, and then via
! some post processing code, one could reconstruct the entire array.
! To keep this code self contained we gather the entire array onto
! processor 0 for final output (say for graphics).


!   Illustration on a 2x2 procr grid
!
!         col 0      col 1
!   ---------------------------
!   |            |            |
!   |   P(0,0)   |   P(0,1)   |  row 0
!   |            |            |
!   ---------------------------
!   |            |            |
!   |   P(1,0)   |   P(1,1)   |  row 1
!   |            |            |
!   ---------------------------

! The final output occurs in two stages.
! (1) in row_comm, MPI_GATHERV the partial rows, so that at the end 
!     of the gatherv, the root procrs in each row has complete row
!    (yet stil partial columns)
!
! (2) next, transpose the matrix. now in col_comm, col 0 only, MPI_GATHERV 
!     the partial rows (which are really the partial columns in step(1)) 
!     so that at the end of the gatherv, the root procr will have the complete matrix.
!
! (3) at the end of these two steps, P(0,0) will have the complete
!     matrix in transpose form

!-------------------------------------------------------------------------------------------
! Notes on the dummy allocation steps (see comments '!dummy allocation' in lines below)
!-------------------------------------------------------------------------------------------
! The arrays recvbuf1 and Tphi_global in Step(1) and Step(2) below only really need to 
! allocated to their appropriate global sizes by the processors that actually use them as 
! receiving buffers. So then why do we have the 'else' part of the if statements below that 
! allocate dummy memory (i.e. 1x1 arrays) for the processors that are NOT on the receiving 
! end of the MPI_GATHERV? The answer is as follows: recvbuf1 and Tphi_global are allocatable
! arrays in F90. As such, they must be allocated to something (here a dummy 1x1 array that
! is never used) for it to be able to particpate in the MPI_GATHERV statement. Without
! this dummy allocation, some compilers (such as the NAG f95 compiler) will complain (and abort)
! that  recvbuf1 and Tphi_global have not been allocated for the processors which areonly sending
! data in the MPI_GATHERV routine, even though the MPI manual states that all receive buffers 
! are ignored for non-receiving processors. On more forgiving compilers (such as IBM's AIX)
! this subterfuge is not necssary, and the 'else' portion of the if statements below
! which allocate mememory for recvbuf1 and Tphi_global can be deleted. To be safe (and portable)
! we allocate recvbuf1 and Tphi_global (to something) on ALL processors.


!-------------------------------------------------------------------------------------------
! Step (1)
!-------------------------------------------------------------------------------------------
	allocate( recvcounts1( np_cols ), displs1( np_cols ) )
	if( coords(2)==0) then		!i.e. col 0 of procr grid, the recving procrs P(i,0)
	    !allocate( recvcounts1( np_cols ), displs1( np_cols ) )
	    allocate( recvbuf1( nmbr_local_rows(coords(1)+1) , M  ) )
	    allocate( recvbuf1_rho( nmbr_local_rows(coords(1)+1) , M  ) )
	    do j = 1, np_cols
               recvcounts1(j) = nmbr_local_rows( coords(1)+1 )*nmbr_local_cols(j)
	    enddo
	    displs1=0
	    do j = 2, np_cols
	       displs1(j) = displs1(j-1) + recvcounts1(j-1)
	    enddo
        else
            allocate( recvbuf1( 1,1 ) )      !dummy allocation for non-receiving procrs
	    allocate( recvbuf1_rho( 1,1 ) )  !dummy allocation for non-receiving procrs
	endif	

	call MPI_GATHERV( phi_local(1:N_local,1:M_local), N_local*M_local, MPI_DOUBLE_PRECISION, &
			  recvbuf1,  recvcounts1, displs1, MPI_DOUBLE_PRECISION,		 & 
			  0, row_comm, ierr )

	call MPI_GATHERV( rho_local(1:N_local,1:M_local), N_local*M_local, MPI_DOUBLE_PRECISION, &
			  recvbuf1_rho,  recvcounts1, displs1, MPI_DOUBLE_PRECISION,		 & 
			  0, row_comm, ierr )

!-------------------------------------------------------------------------------------------
! Step (2)
!-------------------------------------------------------------------------------------------

	if( coords(2)==0 ) then			!i.e. only gatherv in col 0 of procr grid

	    allocate( Trecvbuf1( M, nmbr_local_rows(coords(1)+1) ) )
	    Trecvbuf1 = transpose( recvbuf1 )	!tranpose the matrix recvd above

	    allocate( Trecvbuf1_rho( M, nmbr_local_rows(coords(1)+1) ) )
	    Trecvbuf1_rho = transpose( recvbuf1_rho )	!tranpose the matrix recvd above
 
            allocate( recvcounts2( np_rows ), displs2( np_rows ) )
	    if( coords(1)==0 ) then		!i.e. only P(0,0), the recving end
                !allocate( recvcounts2( np_rows ), displs2( np_rows ) )
	        allocate( Tphi_global(M,N) )	!tranpose of phi_global for output of result
	        allocate( Tphi_global_rho(M,N) )!tranpose of phi_global for output of result
	        do i = 1, np_rows
	           recvcounts2(i) = nmbr_local_rows(coords(1)+1) * M 
	        enddo
	        displs2 = 0
	        do i = 2, np_rows
	           displs2(i) = displs2(i-1) + recvcounts2(i-1)
	        enddo
            else
	        allocate( Tphi_global(1,1) )	!dummy allocation for non-receiving procrs
	        allocate( Tphi_global_rho(1,1) )!dummy allocation for non-receiving procrs     
	    endif


	    call MPI_GATHERV( Trecvbuf1, nmbr_local_rows(coords(1)+1) * M, MPI_DOUBLE_PRECISION,     &
			      Tphi_global, recvcounts2, displs2, MPI_DOUBLE_PRECISION,		     &
			      0, col_comm, ierr )

	    call MPI_GATHERV( Trecvbuf1_rho, nmbr_local_rows(coords(1)+1) * M, MPI_DOUBLE_PRECISION, &
			      Tphi_global_rho, recvcounts2, displs2, MPI_DOUBLE_PRECISION,	     &
			      0, col_comm, ierr )


! Note: if you want the output 'poisson.dat.***' to come out with the iteration number on
!       which the run converged, uncomment the 2nd 'call output' and comment out the first one.
!       As it is called below, the output is named 'poisson.dat.0' for use with gnuplot.
!       Just type gnuplot and at the prompt (gnuplot>), type   load "gnuplot_poisson" (the
!       double quoates are needed).

	    if( coords(1)==0 ) then 							!i.e. only P(0,0)
	        call output( transpose( Tphi_global ),     filename_head,      0) 	!output phi_global
!	        call output( transpose( Tphi_global ),     filename_head,     it) 	!output phi_global
!	        call output( transpose( Tphi_global_rho ), filename_head_rho,  0) 	!output rho_global
		!  also have P(0,0) write out rho_global in binary form
		open(1,file='rho_parallel.bin',form='unformatted',access='direct',recl=4*N*M)
        	write(1,rec=1) real( transpose( Tphi_global_rho/f ) )
        	close(1)
		!  also have P(0,0) write out phi_global in binary form
		open(1,file='phi_parallel.bin',form='unformatted',access='direct',recl=4*N*M)
        	write(1,rec=1) real( transpose( Tphi_global ) )
        	close(1)
	    endif

	endif

!-----------------------------------------------------------------------------
!  diagnostics
!  each procr writes outs its rho_local array in binary form
!-----------------------------------------------------------------------------
!	open(1,file='rho.'//char(48+coords(1))//char(48+coords(2)), &
!               form='unformatted',access='direct',recl=4*N_local*M_local)	
!	write(1,rec=1) real( rho_local(1:N_local,1:M_local) )
!	close(1)

!  diagnostics
!  each procr writes outs its phi_local array in binary form

!	open(1,file='phi.'//char(48+coords(1))//char(48+coords(2)), &
!               form='unformatted',access='direct',recl=4*N_local*M_local)	
!	write(1,rec=1) real( phi_local(1:N_local,1:M_local) )
!	close(1)

!-----------------------------------------------------------------------------
!				Clean up and exit
!-----------------------------------------------------------------------------	
!  end of program
 
	call MPI_BARRIER( MPI_COMM_WORLD, ierr )

	print'(/,a,I2,a)', 'P(',rank,'): saying good-bye'
	if(rank==0) print'(/,a)','END OF PROGRAM'

	call MPI_FINALIZE( ierr )        

      	end


